@@ -9,7 +9,7 @@
  * it under the terms of the MIT license. See LICENSE for details.
  */
 #ifndef LIB_UTIL_FILE_H
-#define LIB_UITL_FILE_H
+#define LIB_UTIL_FILE_H
 
 int util_file_read_line(char *str, size_t size, const char *fmt, ...);
 int util_file_read_l(long *val, int base, const char *fmt, ...);
