@@ -28,12 +28,14 @@ enum target_type {
 	TT_FCP,
 	TT_NSS,
 	TT_NODE,
+	TT_NVME,
 };
 
 enum reipl_type {
 	REIPL_FCP,
 	REIPL_CCW,
-	REIPL_NSS
+	REIPL_NSS,
+	REIPL_NVME,
 };
 
 static const char *const usage_chreipl =
@@ -41,6 +43,7 @@ static const char *const usage_chreipl =
 "\n"
 " chreipl [ccw] [-d] <DEVICE> [OPTIONS]\n"
 " chreipl [fcp] [-d] <DEVICE> [-w] <WWPN> [-l] <LUN> [OPTIONS]\n"
+" chreipl nvme  [-i] <FID> [-s] <NSID> [OPTIONS]\n"
 " chreipl [node] <NODE|DIR> [OPTIONS]\n"
 " chreipl nss [-n] <NAME> [OPTIONS]\n"
 " chreipl [-h] [-v]\n"
@@ -48,6 +51,7 @@ static const char *const usage_chreipl =
 "The following re-IPL targets are supported:\n"
 "  ccw      IPL from CCW device\n"
 "  fcp      IPL from FCP device\n"
+"  nvme     IPL from NVME device\n"
 "  nss      IPL from NSS\n"
 "  node     IPL from device specified by device node or directory\n"
 "\n"
@@ -60,6 +64,7 @@ static const char *const usage_chreipl =
 "Options for ccw target:\n"
 "  -d, --device <DEVICE>   Device number of the CCW IPL device\n"
 "  -L, --loadparm <PARM>   Loadparm specification\n"
+"  -c, --clear 0|1         Control if memory is cleared on re-IPL\n"
 "\n"
 "Options for fcp target:\n"
 "  -d, --device <DEVICE>   Device number of the adapter of the FCP IPL device\n"
@@ -67,6 +72,13 @@ static const char *const usage_chreipl =
 "  -w  --wwpn <WWPN>       World Wide Port Name of the FCP IPL device\n"
 "  -b, --bootprog <BPROG>  Bootprog specification\n"
 "  -L, --loadparm <PARM>   Loadparm specification\n"
+"  -c, --clear 0|1         Control if memory is cleared on re-IPL\n"
+"\n"
+"Options for nvme target:\n"
+"  -i, --fid <FUNCTION_ID>   PCI Function ID of NVME IPL device (hex)\n"
+"  -s  --nsid <NAMESPACE_ID> Namespace ID of NVME IPL device (decimal, default 1)\n"
+"  -b, --bootprog <BPROG>    Bootprog specification\n"
+"  -L, --loadparm <PARM>     Loadparm specification\n"
 "\n"
 "Options for nss target:\n"
 "  -n, --name <NAME>       Identifier of the NSS\n"
@@ -84,6 +96,10 @@ static struct locals {
 	char			lun[20];	/* 18 character +0x" */
 	int			lun_set;
 	char			busid[10];	/* Bus ID e.g. 0.0.4711 */
+	int			fid_set;
+	char			fid[FID_MAX_LEN];
+	int			nsid_set;
+	char			nsid[11];	/* 10 decimal chars + null */
 	int			busid_set;
 	char			dev[15];	/* Device (e.g. dasda) */
 	int			dev_set;
@@ -92,10 +108,11 @@ static struct locals {
 	char			bootparms[4096];
 	int			bootparms_set;
 	int			force_set;
-	enum target_type	target_type;	/* CCW, FCP, NSS or NODE */
+	enum target_type	target_type;	/* CCW,FCP,NVME,NSS or NODE */
 	int			target_type_set;
 	int			target_type_auto_mode;
-	enum reipl_type		reipl_type;	/* CCW, FCP, NSS */
+	enum reipl_type		reipl_type;	/* CCW, FCP, NVME, NSS */
+	int			reipl_clear;
 } l;
 
 static void __noreturn print_usage_chreipl_exit(void)
@@ -226,6 +243,34 @@ static void set_wwpn(const char *wwpn)
 	l.wwpn_set = 1;
 }
 
+static void set_nvme_nsid(const char *nsid)
+{
+	unsigned long long nsid_tmp;
+	char *endptr;
+
+	nsid_tmp = strtoull(nsid, &endptr, 10);
+	if (*endptr)
+		ERR_EXIT("NSID \"%s\" is not a decimal number", nsid);
+	snprintf(l.nsid, sizeof(l.nsid), "%08llu", nsid_tmp);
+	l.nsid_set = 1;
+}
+
+static void set_nvme_fid(const char *fid)
+{
+	unsigned long long fid_tmp;
+	char *endptr;
+
+	fid_tmp = strtoull(fid, &endptr, 16);
+	if (*endptr)
+		ERR_EXIT("FID \"%s\" is not a hexadecimal number", fid);
+	snprintf(l.fid, sizeof(l.fid), "0x%08llx", fid_tmp);
+	l.fid_set = 1;
+
+	/* nsid defaults to 1, if not already set */
+	if (!l.nsid_set)
+		set_nvme_nsid("1");
+}
+
 static void parse_fcp_args(char *nargv[], int nargc)
 {
 	/*
@@ -245,6 +290,28 @@ static void parse_fcp_args(char *nargv[], int nargc)
 	set_lun(nargv[2]);
 }
 
+static void parse_nvme_args(char *nargv[], int nargc)
+{
+	/*
+	 * we might be called like this:
+	 * chreipl nvme 0x13 1
+	 */
+	if (l.busid_set || l.fid_set || l.nsid_set || l.dev_set)
+		ERR_EXIT("Use either options or positional parameters");
+	if (nargc > 2)
+		ERR_EXIT("Too many arguments specified for \"nvme\" re-IPL "
+			 "type");
+	else if (nargc < 1)
+		ERR_EXIT("The \"nvme\" re-IPL type requires function id, and "
+			 "optional namespace id");
+	set_nvme_fid(nargv[0]);
+
+	if (nargc == 2)
+		set_nvme_nsid(nargv[1]);
+	else
+		set_nvme_nsid("1");
+}
+
 static void parse_ccw_args(char *nargv[], int nargc)
 {
 	/*
@@ -285,6 +352,13 @@ static void dev_from_part(char *dev_name)
 		dev_name[i] = 0;
 }
 
+static void dev_from_part_nvme(char *dev_name)
+{
+	char *delim = strrchr(dev_name, 'p');
+	if (delim)
+		*delim = 0;
+}
+
 static int set_reipl_type(const char *dev_name)
 {
 	if (strncmp(dev_name, "dasd", strlen("dasd")) == 0 ||
@@ -292,11 +366,18 @@ static int set_reipl_type(const char *dev_name)
 		l.reipl_type = REIPL_CCW;
 	else if (strncmp(dev_name, "sd", strlen("sd")) == 0)
 		l.reipl_type = REIPL_FCP;
+	else if (strncmp(dev_name, "nvme", strlen("nvme")) == 0)
+		l.reipl_type = REIPL_NVME;
 	else
 		return -1;
 
 	util_strlcpy(l.dev, dev_name, sizeof(l.dev));
-	dev_from_part(l.dev);
+
+	if (l.reipl_type == REIPL_NVME)
+		dev_from_part_nvme(l.dev);
+	else
+		dev_from_part(l.dev);
+
 	l.dev_set = 1;
 	return 0;
 }
@@ -399,6 +480,9 @@ static void parse_pos_args(char *nargv[], int nargc)
 	case TT_FCP:
 		parse_fcp_args(nargv, nargc);
 		break;
+	case TT_NVME:
+		parse_nvme_args(nargv, nargc);
+		break;
 	case TT_CCW:
 		parse_ccw_args(nargv, nargc);
 		break;
@@ -420,6 +504,14 @@ static void check_fcp_opts(void)
 			 "and LUN");
 }
 
+static void check_nvme_opts(void)
+{
+	if (l.nss_name_set || l.wwpn_set || l.lun_set || l.busid_set)
+		ERR_EXIT("Invalid option for \"nvme\" target specified");
+	if (!(l.fid_set && l.nsid_set))
+		ERR_EXIT("The \"nvme\" target requires FID, and optional NSID");
+}
+
 static void check_ccw_opts(void)
 {
 	if (l.bootprog_set || l.lun_set || l.wwpn_set || l.nss_name_set)
@@ -460,6 +552,16 @@ static void set_target_type_auto(const char *arg)
 	}
 }
 
+static void set_reipl_clear(const char *arg)
+{
+	if (arg[0] == '1')
+		l.reipl_clear = 1;
+	else if (arg[0] == '0')
+		l.reipl_clear = 0;
+	else
+		ERR_EXIT("re-IPL clear argument must be either 1 or 0");
+}
+
 static void parse_chreipl_options(int argc, char *argv[])
 {
 	int opt, idx;
@@ -469,14 +571,17 @@ static void parse_chreipl_options(int argc, char *argv[])
 		{ "device",	 required_argument,	NULL, 'd' },
 		{ "lun",	 required_argument,	NULL, 'l' },
 		{ "wwpn",	 required_argument,	NULL, 'w' },
+		{ "fid",	 required_argument,	NULL, 'i' },
+		{ "nsid",	 required_argument,	NULL, 's' },
 		{ "loadparm",	 required_argument,	NULL, 'L' },
 		{ "name",	 required_argument,	NULL, 'n' },
 		{ "bootparms",	 required_argument,	NULL, 'p' },
 		{ "force",	 no_argument,		NULL, 'f' },
 		{ "version",	 no_argument,		NULL, 'v' },
+		{ "clear",	 required_argument,	NULL, 'c' },
 		{ NULL,		 0,			NULL,  0  }
 	};
-	static const char optstr[] = "hcd:vw:l:fL:b:n:p:";
+	static const char optstr[] = "hd:vw:l:fL:b:n:p:c:i:s:";
 
 	/* dont run without any argument */
 	if (argc == 1)
@@ -488,11 +593,15 @@ static void parse_chreipl_options(int argc, char *argv[])
 		set_target_type(TT_CCW, 0);
 	else if (strcmp(argv[1], "nss") == 0)
 		set_target_type(TT_NSS, 0);
+	else if (strcmp(argv[1], "nvme") == 0)
+		set_target_type(TT_NVME, 0);
 	else if (strcmp(argv[1], "node") == 0)
 		set_target_type(TT_NODE, 0);
 	else
 		set_target_type_auto(argv[1]);
 
+	l.reipl_clear = -1;
+
 	while ((opt = getopt_long(argc, argv, optstr, long_opts, &idx)) != -1) {
 		switch (opt) {
 		case 'h':
@@ -500,9 +609,15 @@ static void parse_chreipl_options(int argc, char *argv[])
 		case 'd':
 			set_device(optarg);
 			break;
+		case 'i':
+			set_nvme_fid(optarg);
+			break;
 		case 'l':
 			set_lun(optarg);
 			break;
+		case 's':
+			set_nvme_nsid(optarg);
+			break;
 		case 'w':
 			set_wwpn(optarg);
 			break;
@@ -521,6 +636,9 @@ static void parse_chreipl_options(int argc, char *argv[])
 		case 'f':
 			l.force_set = 1;
 			break;
+		case 'c':
+			set_reipl_clear(optarg);
+			break;
 		case 'v':
 			print_version_exit();
 		default:
@@ -600,6 +718,11 @@ static void chreipl_ccw(void)
 			 strlen(l.bootparms), BOOTPARMS_CCW_MAX);
 	}
 
+	if (l.reipl_clear >= 0) {
+		check_exists("reipl/ccw/clear", "CCW re-IPL clear attribute");
+		write_str(l.reipl_clear ? "1" : "0", "reipl/ccw/clear");
+	}
+
 	/*
 	 * On old systems that use CCW reipl loadparm cannot be set
 	 */
@@ -627,6 +750,12 @@ static void chreipl_fcp(void)
 		ERR_EXIT("Maximum boot parameter length exceeded (%zu/%u)",
 			 strlen(l.bootparms), BOOTPARMS_FCP_MAX);
 	}
+
+	if (l.reipl_clear >= 0) {
+		check_exists("reipl/fcp/clear", "FCP re-IPL clear attribute");
+		write_str(l.reipl_clear ? "1" : "0", "reipl/fcp/clear");
+	}
+
 	/*
 	 * On old systems the FCP reipl loadparm cannot be set
 	 */
@@ -650,6 +779,40 @@ static void chreipl_fcp(void)
 	print_fcp(0, 0);
 }
 
+static void chreipl_nvme(void)
+{
+	check_nvme_opts();
+
+	if (!nvme_is_device(l.fid, l.nsid) && !l.force_set) {
+		ERR_EXIT("Could not find NVME device with fid %s and nsid %s",
+			l.fid, l.nsid);
+	}
+	check_exists("reipl/nvme/fid", "\"nvme\" re-IPL target");
+
+	if (l.bootparms_set && strlen(l.bootparms) > BOOTPARMS_FCP_MAX) {
+		ERR_EXIT("Maximum boot parameter length exceeded (%zu/%u)",
+			 strlen(l.bootparms), BOOTPARMS_FCP_MAX);
+	}
+
+	write_str_optional(l.loadparm, "reipl/nvme/loadparm", l.loadparm_set,
+			   "loadparm");
+	write_str_optional(l.bootparms, "reipl/nvme/scp_data", l.bootparms_set,
+			   "boot parameters");
+	write_str(l.fid, "reipl/nvme/fid");
+	write_str(l.nsid, "reipl/nvme/nsid");
+	/*
+	 * set the boot record logical block address. Master boot
+	 * record. It is always 0 for Linux
+	 */
+	write_str("0", "reipl/nvme/br_lba");
+	if (!l.bootprog_set)
+		sprintf(l.bootprog, "0");
+	write_str(l.bootprog,  "reipl/nvme/bootprog");
+	write_str("nvme", "reipl/reipl_type");
+
+	print_nvme(0, 0);
+}
+
 static void chreipl_nss(void)
 {
 	check_nss_opts();
@@ -690,6 +853,13 @@ static void chreipl_node(void)
 		l.busid_set = 1;
 		chreipl_fcp();
 		break;
+	case REIPL_NVME:
+		nvme_fid_get(l.dev, l.fid);
+		l.fid_set = 1;
+		nvme_nsid_get(l.dev, l.nsid);
+		l.nsid_set = 1;
+		chreipl_nvme();
+		break;
 	default:
 		ERR_EXIT("Internal error: chreipl_node");
 	}
@@ -705,6 +875,9 @@ void cmd_chreipl(int argc, char *argv[])
 	case TT_FCP:
 		chreipl_fcp();
 		break;
+	case TT_NVME:
+		chreipl_nvme();
+		break;
 	case TT_NSS:
 		chreipl_nss();
 		break;
