@@ -1567,15 +1567,17 @@ scan_check(struct scan_token* scan)
 /*
  * Check if kernel and initrd image paths provided by BLS files are readable.
  * If not, add value of 'scan_keyword_target' into search path and silently
- * update scan list.
+ * update scan list if the file exists.
+ * In case neither path works the scan_check code will correctly handle missing
+ * files
  */
-int
-scan_check_bls(struct scan_token *scan)
+void scan_update_bls_path(struct scan_token *scan)
 {
-	int i, rc;
 	char *target_value = NULL;
 	char *img_value = NULL;
-	char *buffer = NULL;
+	char *file = NULL;
+	char *tmp, *value;
+	int i;
 	/*
 	 * In the BLS case, each BLS section heading inherits a keyword
 	 * assignment target= from zipl.conf, and they are all the same.
@@ -1589,38 +1591,46 @@ scan_check_bls(struct scan_token *scan)
 		}
 	}
 	if (!target_value)
-		return -1;
+		return;
 	for (i = 0 ; scan[i].id != scan_id_empty; i++) {
 		if (scan[i].id != scan_id_keyword_assignment)
 			continue;
 		if (scan[i].content.keyword.keyword == scan_keyword_image ||
 		    scan[i].content.keyword.keyword == scan_keyword_ramdisk) {
 
-			rc = misc_check_readable_file(
-				scan[i].content.keyword.value);
-			if (rc) {
+			value = scan[i].content.keyword.value;
+			/*
+			 * put the filename only into the file var before
+			 * checking its presence
+			 */
+			if (contains_address(value)) {
+				tmp = strrchr(value, ',');
+				file = strndup(value, tmp - value);
+			} else {
+				file = value;
+			}
+			if (misc_check_readable_file(file)) {
 				misc_asprintf(&img_value, "%s%s",
-					      target_value,
-					      scan[i].content.keyword.value);
-				rc = misc_check_readable_file(img_value);
-				if (rc) {
-					error_text(
-						"Image file '%s' is not accessible",
-						scan[i].content.keyword.value);
-					return rc;
+					      target_value, file);
+				if (misc_check_readable_file(img_value))
+					continue;
+
+				/*
+				 * when file has stripped the load address part,
+				 * do generate a prefixed value
+				 */
+				if (file != value) {
+					free(file);
+					free(img_value);
+					misc_asprintf(&img_value, "%s%s",
+						      target_value, value);
 				}
-				buffer = (char *)
-					misc_malloc(strlen(img_value) + 1);
-				if (buffer == NULL)
-					return -1;
-				memcpy(buffer, img_value, strlen(img_value));
-				buffer[strlen(img_value)] = 0;
 				free(scan[i].content.keyword.value);
-				scan[i].content.keyword.value = buffer;
+				scan[i].content.keyword.value = img_value;
 			}
 		}
 	}
-	return 0;
+	return;
 }
 
 static int
