@@ -25,6 +25,8 @@
 #include "lib/util_rec.h"
 #include "lib/util_base.h"
 
+ #include <openssl/crypto.h>
+
 #include "utils.h"
 #include "properties.h"
 
@@ -34,14 +36,16 @@
 					} while (0)
 
 /**
- * Checks if the specified card is of type CCA and is online
+ * Checks if the specified card is of the specified type and is online
  *
  * @param[in] card      card number
+ * @param[in] cardtype  card type (CCA, EP11 or ANY)
  *
- * @returns 1 if its a CCA card and is online, 0 if offline and -1 if its
- *          not a CCA card.
+ * @returns 1 if its card of the specified type and is online,
+ *          0 if offline,
+ *          -1 if its not the specified type.
  */
-int sysfs_is_card_online(int card)
+int sysfs_is_card_online(int card, enum card_type cardtype)
 {
 	long int online;
 	char *dev_path;
@@ -69,9 +73,21 @@ int sysfs_is_card_online(int card)
 		rc = 0;
 		goto out;
 	}
-	if (type[4] != 'C') {
-		rc = -1;
-		goto out;
+	switch (cardtype) {
+	case CARD_TYPE_CCA:
+		if (type[4] != 'C') {
+			rc = -1;
+			goto out;
+		}
+		break;
+	case CARD_TYPE_EP11:
+		if (type[4] != 'P') {
+			rc = -1;
+			goto out;
+		}
+		break;
+	default:
+		break;
 	}
 
 out:
@@ -80,21 +96,23 @@ out:
 }
 
 /**
- * Checks if the specified APQN is of type CCA and is online
+ * Checks if the specified APQN is of the specified type and is online
  *
  * @param[in] card      card number
  * @param[in] domain    the domain
+ * @param[in] cardtype  card type (CCA, EP11 or ANY)
  *
- * @returns 1 if its a CCA card and is online, 0 if offline and -1 if its
- *          not a CCA card.
+ * @returns 1 if its card of the specified type and is online,
+ *          0 if offline,
+ *          -1 if its not the specified type.
  */
-int sysfs_is_apqn_online(int card, int domain)
+int sysfs_is_apqn_online(int card, int domain, enum card_type cardtype)
 {
 	long int online;
 	char *dev_path;
 	int rc = 1;
 
-	rc = sysfs_is_card_online(card);
+	rc = sysfs_is_card_online(card, cardtype);
 	if (rc != 1)
 		return rc;
 
@@ -145,7 +163,7 @@ int sysfs_get_card_level(int card)
 		rc = -1;
 		goto out;
 	}
-	if (type[4] != 'C') {
+	if (type[4] != 'C' && type[4] != 'P') {
 		rc = -1;
 		goto out;
 	}
@@ -162,18 +180,62 @@ out:
 }
 
 /**
- * Gets the 8 character ASCII serial number string of an card from the sysfs.
+ * Returns the type of the card. For a CEXnC CARD_TYPE_CCA is returned,
+ * for a CEXnP CARD_TYPE_EP11.
  *
  * @param[in] card      card number
- * @param[out] serialnr Result buffer
+ *
+ * @returns The card type, or -1 of the type can not be determined.
+ */
+enum card_type sysfs_get_card_type(int card)
+{
+	char *dev_path;
+	char type[20];
+	enum card_type cardtype;
+
+	dev_path = util_path_sysfs("bus/ap/devices/card%02x", card);
+	if (!util_path_is_dir(dev_path)) {
+		cardtype = -1;
+		goto out;
+	}
+	if (util_file_read_line(type, sizeof(type), "%s/type", dev_path) != 0) {
+		cardtype = -1;
+		goto out;
+	}
+	if (strncmp(type, "CEX", 3) != 0 || strlen(type) < 5) {
+		cardtype = -1;
+		goto out;
+	}
+	switch (type[4]) {
+	case 'C':
+		cardtype = CARD_TYPE_CCA;
+		break;
+	case 'P':
+		cardtype = CARD_TYPE_EP11;
+		break;
+	default:
+		cardtype = -1;
+		break;
+	}
+
+out:
+	free(dev_path);
+	return cardtype;
+}
+
+/**
+ * Gets the 8-16 character ASCII serial number string of an card from the sysfs.
+ *
+ * @param[in] card      card number
+ * @param[out] serialnr Result buffer. Must be at least SERIALNR_LENGTH long.
  * @param[in] verbose   if true, verbose messages are printed
  *
  * @returns 0 if the serial number was returned. -ENODEV if the APQN is not
- *          available, or is not a CCA card. -ENOTSUP if the serialnr sysfs
- *          attribute is not available, because the zcrypt kernel module is
- *          on an older level.
+ *          available, or is not a CCA or EP11 card.
+ *          -ENOTSUP if the serialnr sysfs attribute is not available, because
+ *          the zcrypt kernel module is on an older level.
  */
-int sysfs_get_serialnr(int card, char serialnr[9], bool verbose)
+int sysfs_get_serialnr(int card, char *serialnr, bool verbose)
 {
 	char *dev_path;
 	int rc = 0;
@@ -181,7 +243,7 @@ int sysfs_get_serialnr(int card, char serialnr[9], bool verbose)
 	if (serialnr == NULL)
 		return -EINVAL;
 
-	if (sysfs_is_card_online(card) != 1)
+	if (sysfs_is_card_online(card, CARD_TYPE_ANY) != 1)
 		return -ENODEV;
 
 	dev_path = util_path_sysfs("bus/ap/devices/card%02x", card);
@@ -189,7 +251,8 @@ int sysfs_get_serialnr(int card, char serialnr[9], bool verbose)
 		rc = -ENODEV;
 		goto out;
 	}
-	if (util_file_read_line(serialnr, 9, "%s/serialnr", dev_path) != 0) {
+	if (util_file_read_line(serialnr, SERIALNR_LENGTH, "%s/serialnr",
+				dev_path) != 0) {
 		rc = -ENOTSUP;
 		goto out;
 	}
@@ -209,11 +272,79 @@ out:
 	return rc;
 }
 
-static int parse_mk_info(char *line, struct mk_info *mk_info)
+/**
+ * Gets the firmware version of an card from the sysfs.
+ * Currently only EP11 cards provide this information.
+ *
+ * @param[in] card      card number
+ * @param[out] fw_version On return: The firmware version numbers
+ * @param[in] verbose   if true, verbose messages are printed
+ *
+ * @returns 0 if the firmware version was returned. -ENODEV if the APQN is not
+ *          available, or is not a CCA or EP11 card.
+ *          -ENOTSUP if the fw_version sysfs attribute is not available, because
+ *          the zcrypt kernel module is on an older level, or because the card
+ *          type does not provide this information.
+ */
+int sysfs_get_firmware_version(int card, struct fw_version *fw_version,
+			       bool verbose)
+{
+	char *dev_path;
+	char buf[50];
+	int rc = 0;
+
+	if (fw_version == NULL)
+		return -EINVAL;
+
+	if (sysfs_is_card_online(card, CARD_TYPE_ANY) != 1)
+		return -ENODEV;
+
+	dev_path = util_path_sysfs("bus/ap/devices/card%02x", card);
+	if (!util_path_is_dir(dev_path)) {
+		rc = -ENODEV;
+		goto out;
+	}
+	if (util_file_read_line(buf, sizeof(buf), "%s/FW_version",
+				dev_path) != 0) {
+		rc = -ENOTSUP;
+		goto out;
+	}
+
+	if (sscanf(buf, "%d.%d", &fw_version->major, &fw_version->minor) != 2) {
+		rc = -ENODEV;
+		goto out;
+	}
+
+	if (util_file_read_line(buf, sizeof(buf), "%s/API_ordinalnr",
+				dev_path) != 0) {
+		rc = -ENOTSUP;
+		goto out;
+	}
+
+	if (sscanf(buf, "%d", &fw_version->api_ordinal) != 1) {
+		rc = -ENODEV;
+		goto out;
+	}
+
+
+	pr_verbose(verbose, "Firmware version of %02x: %d.%d (API: %d)", card,
+		   fw_version->major, fw_version->minor,
+		   fw_version->api_ordinal);
+out:
+	if (rc != 0)
+		pr_verbose(verbose, "Failed to get firmware version for "
+			   "%02x: %s", card, strerror(-rc));
+
+	free(dev_path);
+	return rc;
+}
+
+static int parse_cca_mk_info(char *line, struct mk_info *mk_info)
 {
 	struct mk_info_reg *mk_reg;
 	char *save;
 	char *tok;
+	u64 mkvp;
 
 	tok = strtok_r(line, " ", &save);
 	if (tok == NULL)
@@ -256,9 +387,79 @@ static int parse_mk_info(char *line, struct mk_info *mk_info)
 	if (tok == NULL)
 		return -EIO;
 
-	if (sscanf(tok, "%llx", &mk_reg->mkvp) != 1)
+	if (sscanf(tok, "%llx", &mkvp) != 1)
 		return -EIO;
 
+	memcpy(mk_reg->mkvp, &mkvp, sizeof(mkvp));
+
+	return 0;
+}
+
+static int parse_ep11_mk_info(char *line, struct mk_info *mk_info)
+{
+	struct mk_info_reg *mk_reg;
+	unsigned char *buf;
+	char *save;
+	char *tok;
+	long len;
+
+	tok = strtok_r(line, " ", &save);
+	if (tok == NULL)
+		return -EIO;
+
+	if (strcasecmp(tok, "WK") != 0)
+		return 0;
+
+	tok = strtok_r(NULL, " ", &save);
+	if (tok == NULL)
+		return -EIO;
+
+	if (strcasecmp(tok, "NEW:") == 0)
+		mk_reg = &mk_info->new_mk;
+	else if (strcasecmp(tok, "CUR:") == 0)
+		mk_reg = &mk_info->cur_mk;
+	else
+		return -EIO;
+
+	tok = strtok_r(NULL, " ", &save);
+	if (tok == NULL)
+		return -EIO;
+
+	if (strcasecmp(tok, "valid") == 0)
+		mk_reg->mk_state = MK_STATE_VALID;
+	else if (strcasecmp(tok, "invalid") == 0)
+		mk_reg->mk_state = MK_STATE_INVALID;
+	else if (strcasecmp(tok, "empty") == 0)
+		mk_reg->mk_state = MK_STATE_EMPTY;
+	else if (strcasecmp(tok, "uncommitted") == 0)
+		mk_reg->mk_state = MK_STATE_UNCOMMITTED;
+	else if (strcasecmp(tok, "committed") == 0)
+		mk_reg->mk_state = MK_STATE_COMMITTED;
+	else
+		mk_reg->mk_state = MK_STATE_UNKNOWN;
+
+	tok = strtok_r(NULL, " ", &save);
+	if (tok == NULL)
+		return -EIO;
+
+	/*
+	 * EP11 uses a 32 byte master key verification pattern.
+	 * Usually only the first 16 bytes are used, so we store only up to
+	 * 16 bytes.
+	 */
+	if (strlen(tok) >= MKVP_LENGTH * 2) {
+		if (strncmp(tok, "0x", 2) == 0)
+			tok += 2;
+
+		buf = OPENSSL_hexstr2buf(tok, &len);
+		if (buf == NULL)
+			return -EIO;
+		if (len > MKVP_LENGTH)
+			len = MKVP_LENGTH;
+		memcpy(mk_reg->mkvp, buf, len);
+		OPENSSL_free(buf);
+	}
+
 	return 0;
 }
 
@@ -272,12 +473,13 @@ static int parse_mk_info(char *line, struct mk_info *mk_info)
  * @param[in] verbose   if true, verbose messages are printed
  *
  * @returns 0 if the master key info was returned. -ENODEV if the APQN is not
- *          available, or is not a CCA card. -ENOTSUP if the mkvps sysfs
- *          attribute is not available, because the zcrypt kernel module is
- *          on an older level.
+ *          available, or is not a CCA or EP11 card.
+ *          -ENOTSUP if the mkvps sysfs attribute is not available, because the
+ *          zcrypt kernel module is on an older level.
  */
 int sysfs_get_mkvps(int card, int domain, struct mk_info *mk_info, bool verbose)
 {
+	enum card_type cardtype;
 	char *dev_path;
 	char *p, *end;
 	char buf[100];
@@ -292,9 +494,11 @@ int sysfs_get_mkvps(int card, int domain, struct mk_info *mk_info, bool verbose)
 	mk_info->cur_mk.mk_state = MK_STATE_UNKNOWN;
 	mk_info->old_mk.mk_state = MK_STATE_UNKNOWN;
 
-	if (sysfs_is_apqn_online(card, domain) != 1)
+	if (sysfs_is_apqn_online(card, domain, CARD_TYPE_ANY) != 1)
 		return -ENODEV;
 
+	cardtype = sysfs_get_card_type(card);
+
 	dev_path = util_path_sysfs("bus/ap/devices/card%02x/%02x.%04x/mkvps",
 				   card, card, domain);
 	if (!util_path_is_reg_file(dev_path)) {
@@ -310,14 +514,22 @@ int sysfs_get_mkvps(int card, int domain, struct mk_info *mk_info, bool verbose)
 
 	/*
 	 * Expected contents:
-	 *   AES NEW: <new_mk_state> <new_mk_mkvp>
-	 *   AES CUR: <cur_mk_state> <cur_mk_mkvp>
-	 *   AES OLD: <old_mk_state> <old_mk_mkvp>
-	 * with
-	 *   <new_mk_state>: 'empty' or 'partial' or 'full'
-	 *   <cur_mk_state>, <old_mk_state>: 'valid' or 'invalid'
-	 *   <new_mk_mkvp>, <cur_mk_mkvp>, <old_mk_mkvp:
-	 *        8 byte hex string with leading 0x
+	 * For CCA cards:
+	 *     AES NEW: <new_mk_state> <new_mk_mkvp>
+	 *     AES CUR: <cur_mk_state> <cur_mk_mkvp>
+	 *     AES OLD: <old_mk_state> <old_mk_mkvp>
+	 *   with
+	 *     <new_mk_state>: 'empty' or 'partial' or 'full'
+	 *     <cur_mk_state>, <old_mk_state>: 'valid' or 'invalid'
+	 *     <new_mk_mkvp>, <cur_mk_mkvp>, <old_mk_mkvp>:
+	 *          8 byte hex string with leading 0x
+	 * For EP11 cards:
+	 *     WK NEW: <new_wk_state> <new_wk_mkvp>
+	 *     WK CUR: <cur_wk_state> <cur_wk_mkvp>
+	 *   with
+	 *     <wk_cur_state>: 'invalid' or 'valid'
+	 *     <wk_new_state>: 'empty' or 'uncommitted' or 'committed'
+	 *     <wk_cur_vp> and <wk_new_vp>: '-' or a 32 byte hash pattern
 	 */
 	while ((p = fgets(buf, sizeof(buf), fp)) != NULL) {
 		end = memchr(buf, '\n', sizeof(buf));
@@ -329,7 +541,17 @@ int sysfs_get_mkvps(int card, int domain, struct mk_info *mk_info, bool verbose)
 		pr_verbose(verbose, "mkvp for %02x.%04x: %s", card, domain,
 			   buf);
 
-		rc = parse_mk_info(buf, mk_info);
+		switch (cardtype) {
+		case CARD_TYPE_CCA:
+			rc = parse_cca_mk_info(buf, mk_info);
+			break;
+		case CARD_TYPE_EP11:
+			rc = parse_ep11_mk_info(buf, mk_info);
+			break;
+		default:
+			rc = -EINVAL;
+			break;
+		}
 		if (rc != 0)
 			break;
 	}
@@ -338,7 +560,8 @@ int sysfs_get_mkvps(int card, int domain, struct mk_info *mk_info, bool verbose)
 
 	if (mk_info->new_mk.mk_state == MK_STATE_UNKNOWN &&
 	    mk_info->cur_mk.mk_state == MK_STATE_UNKNOWN &&
-	    mk_info->old_mk.mk_state == MK_STATE_UNKNOWN)
+	    (cardtype == CARD_TYPE_CCA &&
+	     mk_info->old_mk.mk_state == MK_STATE_UNKNOWN))
 		rc = -EIO;
 out:
 	if (rc != 0)
@@ -349,8 +572,9 @@ out:
 	return rc;
 }
 
-static int scan_for_domains(int card, apqn_handler_t handler,
-			    void *handler_data, bool verbose)
+static int scan_for_domains(int card, enum card_type cardtype,
+			    apqn_handler_t handler, void *handler_data,
+			    bool verbose)
 {
 	struct dirent **namelist;
 	char fname[290];
@@ -369,9 +593,9 @@ static int scan_for_domains(int card, apqn_handler_t handler,
 
 		pr_verbose(verbose, "Found %02x.%04x", card, domain);
 
-		if (sysfs_is_apqn_online(card, domain) != 1) {
+		if (sysfs_is_apqn_online(card, domain, cardtype) != 1) {
 			pr_verbose(verbose, "APQN %02x.%04x is offline or not "
-				   "CCA", card, domain);
+				   "the correct type", card, domain);
 			continue;
 		}
 
@@ -385,8 +609,8 @@ static int scan_for_domains(int card, apqn_handler_t handler,
 }
 
 
-static int scan_for_apqns(apqn_handler_t handler, void *handler_data,
-			  bool verbose)
+static int scan_for_apqns(enum card_type cardtype, apqn_handler_t handler,
+			  void *handler_data, bool verbose)
 {
 	struct dirent **namelist;
 	int i, n, card, rc = 0;
@@ -405,13 +629,14 @@ static int scan_for_apqns(apqn_handler_t handler, void *handler_data,
 
 		pr_verbose(verbose, "Found card %02x", card);
 
-		if (sysfs_is_card_online(card) != 1) {
-			pr_verbose(verbose, "Card %02x is offline or not CCA",
-				   card);
+		if (sysfs_is_card_online(card, cardtype) != 1) {
+			pr_verbose(verbose, "Card %02x is offline or not the "
+				   "correct type", card);
 			continue;
 		}
 
-		rc = scan_for_domains(card, handler, handler_data, verbose);
+		rc = scan_for_domains(card, cardtype, handler, handler_data,
+				      verbose);
 		if (rc != 0)
 			break;
 	}
@@ -421,21 +646,22 @@ static int scan_for_apqns(apqn_handler_t handler, void *handler_data,
 }
 
 /**
- * Calls the handler for all APQNs specified in the apqns parameter, or of this
- * is NULL, for all online CCA APQNs found in sysfs. In case sysfs is inspected,
- * the cards and domains are processed in alphabetical order.
+ * Calls the handler for all APQNs specified in the apqns parameter, or if this
+ * is NULL, for all online CCA or EP11 APQNs found in sysfs. In case sysfs is
+ * inspected, the cards and domains are processed in alphabetical order.
  *
  * @param[in] apqns     a comma separated list of APQNs. If NULL is specified,
- *                      or an empty string, then all online CCA APQNs are
- *                      handled.
+ *                      or an empty string, then all online CCA or EP11 APQNs
+ *                      are handled.
+ * @param[in] cardtype  card type (CCA, EP11 or ANY)
  * @param[in] handler   a handler function that is called for each APQN
  * @param[in] handler_data private data that is passed to the handler
  * @param[in] verbose   if true, verbose messages are printed
  *
  * @returns 0 for success or a negative errno in case of an error
  */
-int handle_apqns(const char *apqns, apqn_handler_t handler, void *handler_data,
-		 bool verbose)
+int handle_apqns(const char *apqns, enum card_type cardtype,
+		 apqn_handler_t handler, void *handler_data, bool verbose)
 {
 	int card, domain;
 	char *copy, *tok;
@@ -443,7 +669,7 @@ int handle_apqns(const char *apqns, apqn_handler_t handler, void *handler_data,
 	int rc = 0;
 
 	if (apqns == NULL || (apqns != NULL && strlen(apqns) == 0)) {
-		rc = scan_for_apqns(handler, handler_data, verbose);
+		rc = scan_for_apqns(cardtype, handler, handler_data, verbose);
 	} else {
 		copy = util_strdup(apqns);
 		tok = strtok_r(copy, ",", &save);
@@ -472,6 +698,7 @@ int handle_apqns(const char *apqns, apqn_handler_t handler, void *handler_data,
 
 struct print_apqn_info {
 	struct util_rec *rec;
+	enum card_type cardtype;
 	bool verbose;
 };
 
@@ -480,33 +707,41 @@ static int print_apqn_mk_info(int card, int domain, void *handler_data)
 	struct print_apqn_info *info = (struct print_apqn_info *)handler_data;
 	struct mk_info mk_info;
 	int rc, level;
+	enum card_type type;
 
 	rc = sysfs_get_mkvps(card, domain, &mk_info, info->verbose);
 	if (rc == -ENOTSUP)
 		return rc;
 
 	level = sysfs_get_card_level(card);
+	type = sysfs_get_card_type(card);
 
 	util_rec_set(info->rec, "APQN", "%02x.%04x", card, domain);
 
+	if (info->cardtype != CARD_TYPE_ANY && type != info->cardtype)
+		rc = -EINVAL;
+
 	if (rc == 0) {
-		if (mk_info.new_mk.mk_state == MK_STATE_FULL)
-			util_rec_set(info->rec, "NEW", "%016llx",
-				     mk_info.new_mk.mkvp);
+		if (mk_info.new_mk.mk_state == MK_STATE_FULL ||
+		    mk_info.new_mk.mk_state == MK_STATE_COMMITTED)
+			util_rec_set(info->rec, "NEW", "%s",
+				     printable_mkvp(type, mk_info.new_mk.mkvp));
 		else if (mk_info.new_mk.mk_state == MK_STATE_PARTIAL)
 			util_rec_set(info->rec, "NEW", "partially loaded");
+		else if (mk_info.new_mk.mk_state == MK_STATE_UNCOMMITTED)
+			util_rec_set(info->rec, "NEW", "uncommitted");
 		else
 			util_rec_set(info->rec, "NEW", "-");
 
 		if (mk_info.cur_mk.mk_state ==  MK_STATE_VALID)
-			util_rec_set(info->rec, "CUR", "%016llx",
-				     mk_info.cur_mk.mkvp);
+			util_rec_set(info->rec, "CUR", "%s",
+				     printable_mkvp(type, mk_info.cur_mk.mkvp));
 		else
 			util_rec_set(info->rec, "CUR", "-");
 
 		if (mk_info.old_mk.mk_state ==  MK_STATE_VALID)
-			util_rec_set(info->rec, "OLD", "%016llx",
-				     mk_info.old_mk.mkvp);
+			util_rec_set(info->rec, "OLD", "%s",
+				     printable_mkvp(type, mk_info.old_mk.mkvp));
 		else
 			util_rec_set(info->rec, "OLD", "-");
 	} else {
@@ -515,8 +750,9 @@ static int print_apqn_mk_info(int card, int domain, void *handler_data)
 		util_rec_set(info->rec, "OLD", "?");
 	}
 
-	if (level > 0)
-		util_rec_set(info->rec, "TYPE", "CEX%dC", level);
+	if (level > 0 && type != CARD_TYPE_ANY)
+		util_rec_set(info->rec, "TYPE", "CEX%d%c", level,
+			     type == CARD_TYPE_CCA ? 'C' : 'P');
 	else
 		util_rec_set(info->rec, "TYPE", "?");
 
@@ -529,40 +765,51 @@ static int print_apqn_mk_info(int card, int domain, void *handler_data)
  * Prints master key information for all specified APQNs
  *
  * @param[in] apqns     a comma separated list of APQNs. If NULL is specified,
- *                      or an empty string, then all online CCA APQNs are
- *                      printed.
+ *                      or an empty string, then all online CCA or EP11 APQNs
+ *                      are printed.
+ * @param[in] cardtype  card type (CCA, EP11 or ANY)
  * @param[in] verbose   if true, verbose messages are printed
  *
  * @returns 0 for success or a negative errno in case of an error. -ENOTSUP is
  *          returned when the mkvps sysfs attribute is not available, because
  *          the zcrypt kernel module is on an older level.
  */
-int print_mk_info(const char *apqns, bool verbose)
+int print_mk_info(const char *apqns, enum card_type cardtype, bool verbose)
 {
 	struct print_apqn_info info;
-	int rc;
+	int rc, mklen;
 
 	info.verbose = verbose;
+	info.cardtype = cardtype;
 	info.rec = util_rec_new_wide("-");
 
+	if (cardtype == CARD_TYPE_CCA)
+		mklen = 16;
+	else
+		mklen = 32;
+
 	util_rec_def(info.rec, "APQN", UTIL_REC_ALIGN_LEFT, 11, "CARD.DOMAIN");
-	util_rec_def(info.rec, "NEW", UTIL_REC_ALIGN_LEFT, 16, "NEW MK");
-	util_rec_def(info.rec, "CUR", UTIL_REC_ALIGN_LEFT, 16, "CURRENT MK");
-	util_rec_def(info.rec, "OLD", UTIL_REC_ALIGN_LEFT, 16, "OLD MK");
+	util_rec_def(info.rec, "NEW", UTIL_REC_ALIGN_LEFT, mklen, "NEW MK");
+	util_rec_def(info.rec, "CUR", UTIL_REC_ALIGN_LEFT, mklen, "CURRENT MK");
+	if (cardtype != CARD_TYPE_EP11)
+		util_rec_def(info.rec, "OLD", UTIL_REC_ALIGN_LEFT, mklen,
+			     "OLD MK");
 	util_rec_def(info.rec, "TYPE", UTIL_REC_ALIGN_LEFT, 6, "TYPE");
 	util_rec_print_hdr(info.rec);
 
-	rc = handle_apqns(apqns, print_apqn_mk_info, &info, verbose);
+	rc = handle_apqns(apqns, cardtype, print_apqn_mk_info, &info, verbose);
 
 	util_rec_free(info.rec);
 	return rc;
 }
 
 struct cross_check_info {
-	u64	mkvp;
-	u64	new_mkvp;
+	u8	mkvp[MKVP_LENGTH];
+	u8	new_mkvp[MKVP_LENGTH];
 	bool	key_mkvp;
+	enum card_type cardtype;
 	int	min_level;
+	const struct fw_version *min_fw_version;
 	u32	num_cur_match;
 	u32	num_old_match;
 	u32	num_new_match;
@@ -575,15 +822,17 @@ struct cross_check_info {
 static int cross_check_mk_info(int card, int domain, void *handler_data)
 {
 	struct cross_check_info *info = (struct cross_check_info *)handler_data;
+	struct fw_version fw_version;
+	enum card_type type;
 	struct mk_info mk_info;
 	char temp[200];
-	int rc, level;
+	int rc, level = 0;
 
 	rc = sysfs_get_mkvps(card, domain, &mk_info, info->verbose);
 	if (rc == -ENODEV) {
 		info->print_mks = 1;
 		printf("WARNING: APQN %02x.%04x: Not available or not of "
-		       "type CCA\n", card, domain);
+		       "the correct type\n", card, domain);
 		return 0;
 	}
 	if (rc != 0)
@@ -591,6 +840,19 @@ static int cross_check_mk_info(int card, int domain, void *handler_data)
 
 	info->num_checked++;
 
+	if (info->cardtype != CARD_TYPE_ANY) {
+		type = sysfs_get_card_type(card);
+		if (type != info->cardtype) {
+			info->print_mks = 1;
+			info->mismatch = 1;
+			sprintf(temp, "WARNING: APQN %02x.%04x: The card type "
+				"is not CEXn%c.", card, domain,
+				info->cardtype == CARD_TYPE_CCA ? 'C' : 'P');
+			util_print_indented(temp, 0);
+			return 0;
+		}
+	}
+
 	if (info->min_level >= 0) {
 		level = sysfs_get_card_level(card);
 
@@ -598,25 +860,63 @@ static int cross_check_mk_info(int card, int domain, void *handler_data)
 			info->print_mks = 1;
 			info->mismatch = 1;
 			sprintf(temp, "WARNING: APQN %02x.%04x: The card level "
-				"is less than CEX%dC.", card, domain,
+				"is less than CEX%dn.", card, domain,
 				info->min_level);
 			util_print_indented(temp, 0);
 		}
 	}
 
+	if (info->min_fw_version != NULL) {
+		rc = sysfs_get_firmware_version(card, &fw_version,
+					       info->verbose);
+		if (rc == 0) {
+			if (fw_version.api_ordinal <
+					info->min_fw_version->api_ordinal) {
+				info->print_mks = 1;
+				info->mismatch = 1;
+				sprintf(temp, "WARNING: APQN %02x.%04x: The "
+					"firmware version is too less to "
+					"support secure keys of that type",
+					card, domain);
+				util_print_indented(temp, 0);
+			}
+			if (info->min_level > 0 && info->min_level == level &&
+			    (fw_version.major < info->min_fw_version->major ||
+			     (fw_version.major == info->min_fw_version->major &&
+			      fw_version.minor < info->min_fw_version->minor))) {
+				info->print_mks = 1;
+				info->mismatch = 1;
+				sprintf(temp, "WARNING: APQN %02x.%04x: The "
+					"firmware version is too less to "
+					"support secure keys of that type",
+					card, domain);
+				util_print_indented(temp, 0);
+			}
+		}
+	}
+
 	if (mk_info.new_mk.mk_state == MK_STATE_PARTIAL) {
 		info->print_mks = 1;
 		sprintf(temp, "INFO: APQN %02x.%04x: The NEW master key "
 			"register is only partially loaded.", card, domain);
 		util_print_indented(temp, 0);
 	}
+	if (mk_info.new_mk.mk_state == MK_STATE_UNCOMMITTED) {
+		info->print_mks = 1;
+		sprintf(temp, "INFO: APQN %02x.%04x: The NEW master key "
+			"register is loaded but uncommitted.", card, domain);
+		util_print_indented(temp, 0);
+	}
 
-	if (info->new_mkvp == 0 &&
-	    mk_info.new_mk.mk_state == MK_STATE_FULL)
-		info->new_mkvp = mk_info.new_mk.mkvp;
+	if (MKVP_ZERO(info->new_mkvp) &&
+	    (mk_info.new_mk.mk_state == MK_STATE_FULL ||
+	     mk_info.new_mk.mk_state == MK_STATE_COMMITTED))
+		memcpy(info->new_mkvp, mk_info.new_mk.mkvp,
+		       sizeof(info->new_mkvp));
 
-	if (mk_info.new_mk.mk_state == MK_STATE_FULL &&
-	    mk_info.new_mk.mkvp != info->new_mkvp) {
+	if ((mk_info.new_mk.mk_state == MK_STATE_FULL ||
+	     mk_info.new_mk.mk_state == MK_STATE_COMMITTED) &&
+	    !MKVP_EQ(mk_info.new_mk.mkvp, info->new_mkvp)) {
 		info->print_mks = 1;
 		sprintf(temp, "WARNING: APQN %02x.%04x: The NEW master key "
 			      "register contains a different master key than "
@@ -634,15 +934,16 @@ static int cross_check_mk_info(int card, int domain, void *handler_data)
 	}
 
 	if (mk_info.old_mk.mk_state == MK_STATE_VALID &&
-	    mk_info.old_mk.mkvp == mk_info.cur_mk.mkvp) {
+	    MKVP_EQ(mk_info.old_mk.mkvp, mk_info.cur_mk.mkvp)) {
 		info->print_mks = 1;
 		sprintf(temp, "INFO: APQN %02x.%04x: The OLD master key "
 			"register contains the same master key as the CURRENT "
 			"master key register.", card, domain);
 		util_print_indented(temp, 0);
 	}
-	if (mk_info.new_mk.mk_state == MK_STATE_FULL &&
-	    mk_info.new_mk.mkvp == mk_info.cur_mk.mkvp) {
+	if ((mk_info.new_mk.mk_state == MK_STATE_FULL ||
+	     mk_info.new_mk.mk_state == MK_STATE_COMMITTED) &&
+	    MKVP_EQ(mk_info.new_mk.mkvp, mk_info.cur_mk.mkvp)) {
 		info->print_mks = 1;
 		sprintf(temp, "INFO: APQN %02x.%04x: The NEW master key "
 			"register contains the same master key as the CURRENT "
@@ -651,7 +952,7 @@ static int cross_check_mk_info(int card, int domain, void *handler_data)
 	}
 	if (mk_info.new_mk.mk_state == MK_STATE_FULL &&
 	    mk_info.old_mk.mk_state == MK_STATE_VALID &&
-	    mk_info.new_mk.mkvp == mk_info.old_mk.mkvp) {
+	    MKVP_EQ(mk_info.new_mk.mkvp, mk_info.old_mk.mkvp)) {
 		info->print_mks = 1;
 		sprintf(temp, "INFO: APQN %02x.%04x: The NEW master key "
 			"register contains the same master key as the OLD "
@@ -659,28 +960,29 @@ static int cross_check_mk_info(int card, int domain, void *handler_data)
 		util_print_indented(temp, 0);
 	}
 
-	if (info->mkvp == 0)
-		info->mkvp = mk_info.cur_mk.mkvp;
+	if (MKVP_ZERO(info->mkvp))
+		memcpy(info->mkvp, mk_info.cur_mk.mkvp, sizeof(info->mkvp));
 
 	if (info->key_mkvp) {
 		if (mk_info.cur_mk.mk_state == MK_STATE_VALID &&
-		    mk_info.cur_mk.mkvp == info->mkvp)
+		    MKVP_EQ(mk_info.cur_mk.mkvp, info->mkvp))
 			info->num_cur_match++;
 
 		if (mk_info.old_mk.mk_state == MK_STATE_VALID &&
-		    mk_info.old_mk.mkvp == info->mkvp)
+		    MKVP_EQ(mk_info.old_mk.mkvp, info->mkvp))
 			info->num_old_match++;
 
-		if (mk_info.new_mk.mk_state == MK_STATE_FULL &&
-		    mk_info.new_mk.mkvp == info->mkvp)
+		if ((mk_info.new_mk.mk_state == MK_STATE_FULL ||
+		     mk_info.new_mk.mk_state == MK_STATE_COMMITTED) &&
+		    MKVP_EQ(mk_info.new_mk.mkvp, info->mkvp))
 			info->num_new_match++;
 	}
 
-	if (mk_info.cur_mk.mkvp != info->mkvp) {
+	if (!MKVP_EQ(mk_info.cur_mk.mkvp, info->mkvp)) {
 
 		if (info->key_mkvp) {
 			if (mk_info.old_mk.mk_state == MK_STATE_VALID &&
-			    mk_info.old_mk.mkvp == info->mkvp) {
+			    MKVP_EQ(mk_info.old_mk.mkvp, info->mkvp)) {
 				info->print_mks = 1;
 				sprintf(temp, "INFO: APQN %02x.%04x: The master"
 					" key has been changed to a new "
@@ -688,8 +990,10 @@ static int cross_check_mk_info(int card, int domain, void *handler_data)
 					"not yet been re-enciphered.", card,
 					domain);
 				util_print_indented(temp, 0);
-			} else if (mk_info.new_mk.mk_state == MK_STATE_FULL &&
-				   mk_info.new_mk.mkvp == info->mkvp) {
+			} else if ((mk_info.new_mk.mk_state == MK_STATE_FULL ||
+				    mk_info.new_mk.mk_state ==
+							MK_STATE_COMMITTED) &&
+				   MKVP_EQ(mk_info.new_mk.mkvp, info->mkvp)) {
 				info->print_mks = 1;
 				sprintf(temp, "INFO: APQN %02x.%04x: The master"
 					" key has been changed but is not "
@@ -722,13 +1026,16 @@ static int cross_check_mk_info(int card, int domain, void *handler_data)
  * out an information message about the APQNs that have a different master key.
  *
  * @param[in] apqns     a comma separated list of APQNs. If NULL is specified,
- *                      or an empty string, then all online CCA APQNs are
+ *                      or an empty string, then all online APQNs are
  *                      checked.
  * @param[in] mkvp      The master key verification pattern of a secure key.
- *                      If this is all zero, then the master keys are not
- *                      matched against it.
+ *                      If this is all zero or NULL, then the master keys are
+ *                      not matched against it.
  * @param[in] min_level The minimum card level required. If min_level is -1 then
  *                      the card level is not checked.
+ * @param[in] min_fw_version The minimum firmware version required. If NULL tne
+ *                      the firmware version is not checked.
+ * @param[in] cardtype  card type (CCA, EP11 or ANY)
  * @param[in] print_mks if true, then a the full master key info of all
  *                      specified APQns is printed, in case of a mismatch.
  * @param[in] verbose   if true, verbose messages are printed
@@ -738,40 +1045,49 @@ static int cross_check_mk_info(int card, int domain, void *handler_data)
  *          -ENOTSUP is returned when the mkvps sysfs attribute is not
  *          available, because the zcrypt kernel module is on an older level.
  */
-int cross_check_apqns(const char *apqns, u64 mkvp, int min_level,
-		      bool print_mks, bool verbose)
+int cross_check_apqns(const char *apqns, u8 *mkvp, int min_level,
+		      const struct fw_version *min_fw_version,
+		      enum card_type cardtype, bool print_mks, bool verbose)
 {
 	struct cross_check_info info;
 	char temp[200];
 	int rc;
 
 	memset(&info, 0, sizeof(info));
-	info.key_mkvp = mkvp != 0;
-	info.mkvp = mkvp;
+	info.key_mkvp = !MKVP_ZERO(mkvp);
+	if (mkvp != NULL)
+		memcpy(info.mkvp, mkvp, sizeof(info.mkvp));
+	info.cardtype = cardtype;
 	info.min_level = min_level;
+	info.min_fw_version = min_fw_version;
 	info.verbose = verbose;
 
-	pr_verbose(verbose, "Cross checking APQNs with mkvp 0x%016llx and "
-		   "min-level %d: %s", mkvp, min_level,
+	pr_verbose(verbose, "Cross checking APQNs with mkvp %s, "
+		   "min-level %d, and min-fw-version %u.%u (api: %u): %s",
+		   printable_mkvp(cardtype, info.mkvp),
+		   min_level,
+		   min_fw_version != NULL ? min_fw_version->major : 0,
+		   min_fw_version != NULL ? min_fw_version->minor : 0,
+		   min_fw_version != NULL ? min_fw_version->api_ordinal : 0,
 		   apqns != NULL ? apqns : "ANY");
 
-	rc = handle_apqns(apqns, cross_check_mk_info, &info, verbose);
+	rc = handle_apqns(apqns, cardtype, cross_check_mk_info, &info, verbose);
 	if (rc != 0)
 		return rc;
 
 	if (info.mismatch) {
 		if (info.key_mkvp)
 			printf("WARNING: Not all APQNs have the correct master "
-			       "key (%016llx).\n", mkvp);
+			       "key (%s) or fulfill the requirements.\n",
+			       printable_mkvp(cardtype, info.mkvp));
 		else
 			printf("WARNING: Not all APQNs have the same master "
-			       "key.\n");
-
+			       "key or fulfill the requirements.\n");
 		rc = -ENODEV;
 	}
 	if (info.num_checked == 0) {
 		printf("WARNING: None of the APQNs is available or of "
-		       "type CCA\n");
+		       "the correct type\n");
 		rc = -ENODEV;
 	}
 	if (info.num_old_match > 0 && info.num_new_match > 0) {
@@ -787,7 +1103,7 @@ int cross_check_apqns(const char *apqns, u64 mkvp, int min_level,
 
 	if (print_mks && info.print_mks) {
 		printf("\n");
-		print_mk_info(apqns, verbose);
+		print_mk_info(apqns, cardtype, verbose);
 		printf("\n");
 	}
 
@@ -817,3 +1133,38 @@ bool prompt_for_yes(bool verbose)
 
 	return false;
 }
+
+/*
+ * Returns a printable version of the specified master key verification pattern
+ * (MKVP) for the specified card type. Different card types use different
+ * number of bytes for MKVP.
+ *
+ * @param[in] cardtype  card type (CCA, EP11 or ANY)
+ * @param[in] mkvp       the master key verification pattern to print
+ *
+ * @returns address of a static char array containing the printed MKVP, or NULL
+ *          in case of an error.
+ */
+char *printable_mkvp(enum card_type cardtype, u8 *mkvp)
+{
+	static char mkvp_print_buf[MKVP_LENGTH * 2 + 1];
+
+	if (mkvp == NULL)
+		return NULL;
+
+	switch (cardtype) {
+	case CARD_TYPE_CCA:
+		/* CCA uses an 8 byte MKVP */
+		sprintf(mkvp_print_buf, "%016llx", *((u64 *)mkvp));
+		break;
+	case CARD_TYPE_EP11:
+		/* EP11 uses an 32 byte MKVP, but truncated to 16 bytes*/
+		sprintf(mkvp_print_buf, "%016llx%016llx", *((u64 *)&mkvp[0]),
+			*((u64 *)&mkvp[8]));
+		break;
+	default:
+		return NULL;
+	}
+
+	return mkvp_print_buf;
+}
