@@ -28,6 +28,7 @@
 #include "lib/zt_common.h"
 
 #include "cca.h"
+#include "ep11.h"
 #include "keystore.h"
 #include "misc.h"
 #include "pkey.h"
@@ -83,12 +84,16 @@ static struct zkey_globals {
 	bool force;
 	bool open;
 	bool format;
+	struct ext_lib lib;
 	struct cca_lib cca;
+	struct ep11_lib ep11;
 	int pkey_fd;
 	struct keystore *keystore;
 } g = {
 	.pkey_fd = -1,
 	.sector_size = -1,
+	.lib.cca = &g.cca,
+	.lib.ep11 = &g.ep11,
 };
 
 /*
@@ -222,9 +227,9 @@ static struct util_opt opt_vec[] = {
 		.option = { "key-type", required_argument, NULL, 'K'},
 		.argument = "type",
 		.desc = "The type of the key. Possible values are '"
-			KEY_TYPE_CCA_AESDATA"' and '"KEY_TYPE_CCA_AESCIPHER"'. "
-			"When this option is omitted, the default is '"
-			KEY_TYPE_CCA_AESDATA"'",
+			KEY_TYPE_CCA_AESDATA"', '"KEY_TYPE_CCA_AESCIPHER"' "
+			"and '"KEY_TYPE_EP11_AES"'. When this option is "
+			"omitted, the default is '"KEY_TYPE_CCA_AESDATA"'",
 		.command = COMMAND_GENERATE,
 	},
 	/***********************************************************/
@@ -259,7 +264,7 @@ static struct util_opt opt_vec[] = {
 	{
 		.option = {"complete", 0, NULL, 'p'},
 		.desc = "Completes a staged re-enciphering. Use this option "
-			"after the new CCA master key has been set (made "
+			"after the new master key has been set (made "
 			"active)",
 		.command = COMMAND_REENCIPHER,
 	},
@@ -447,9 +452,9 @@ static struct util_opt opt_vec[] = {
 		.option = { "key-type", required_argument, NULL, 'K'},
 		.argument = "type",
 		.desc = "The type of the key. Possible values are '"
-			KEY_TYPE_CCA_AESDATA"' and '"KEY_TYPE_CCA_AESCIPHER"'. "
-			"Use this option to list all keys with the specified "
-			"key type.",
+			KEY_TYPE_CCA_AESDATA"', '"KEY_TYPE_CCA_AESCIPHER"' "
+			"and '"KEY_TYPE_EP11_AES"'. Use this option to list "
+			"all keys with the specified key type.",
 		.command = COMMAND_LIST,
 	},
 	/***********************************************************/
@@ -822,6 +827,7 @@ struct zkey_command {
 	unsigned int abbrev_len;
 	int (*function)(void);
 	int need_cca_library;
+	int need_ep11_library;
 	int need_pkey_device;
 	char *short_desc;
 	char *long_desc;
@@ -868,13 +874,13 @@ static struct zkey_command zkey_commands[] = {
 		.command = COMMAND_REENCIPHER,
 		.abbrev_len = 2,
 		.function = command_reencipher,
-		.need_cca_library = 1,
+		/* Will load the CCA or EP11 library on demand */
 		.need_pkey_device = 1,
 		.short_desc = "Re-encipher an existing secure AES key",
 		.long_desc = "Re-encipher an existing secure AES "
 			     "key that is either contained in SECURE-KEY-FILE "
 			     "or is stored in the repository with another "
-			     "CCA master key",
+			     "master key",
 		.has_options = 1,
 		.pos_arg = "[SECURE-KEY-FILE]",
 		.pos_arg_optional = 1,
@@ -1171,8 +1177,10 @@ static int command_generate(void)
 			return EXIT_FAILURE;
 		}
 
-		rc = cross_check_apqns(NULL, 0,
+		rc = cross_check_apqns(NULL, NULL,
 				get_min_card_level_for_keytype(g.key_type),
+				get_min_fw_version_for_keytype(g.key_type),
+				get_card_type_for_keytype(g.key_type),
 				true, g.verbose);
 		if (rc == -EINVAL)
 			return EXIT_FAILURE;
@@ -1198,9 +1206,8 @@ static int command_reencipher_file(void)
 {
 	size_t secure_key_size;
 	int rc, is_old_mk;
-	int selected = 1;
 	u8 *secure_key;
-	u64 mkvp;
+	bool selected;
 
 	if (g.name != NULL) {
 		warnx("Option '--name|-N' is not valid for "
@@ -1246,30 +1253,21 @@ static int command_reencipher_file(void)
 		goto out;
 	}
 
-	rc = get_master_key_verification_pattern(secure_key, secure_key_size,
-						 &mkvp, g.verbose);
-	if (rc != 0) {
-		warnx("Failed to get the master key verification pattern: %s",
-		      strerror(-rc));
-		rc = EXIT_FAILURE;
-		goto out;
-	}
-
 	if (!g.fromold && !g.tonew) {
 		/* Autodetect reencipher option */
 		if (is_old_mk) {
 			g.fromold = 1;
 			util_print_indented("The secure key is currently "
-					    "enciphered with the OLD CCA "
+					    "enciphered with the OLD "
 					    "master key and is being "
 					    "re-enciphered with the CURRENT "
-					    "CCA master key\n", 0);
+					    "master key\n", 0);
 		} else {
 			g.tonew = 1;
 			util_print_indented("The secure key is currently "
-					    "enciphered with the CURRENT CCA "
+					    "enciphered with the CURRENT "
 					    "master key and is being "
-					    "re-enciphered with the NEW CCA "
+					    "re-enciphered with the NEW "
 					    "master key\n", 0);
 		}
 	}
@@ -1278,68 +1276,56 @@ static int command_reencipher_file(void)
 	if (g.fromold) {
 		if (!is_old_mk) {
 			warnx("The secure key is already enciphered "
-			      "with the CURRENT CCA master key");
+			      "with the CURRENT master key");
 			rc = EXIT_FAILURE;
 			goto out;
 		}
 
 		pr_verbose("Secure key will be re-enciphered from OLD to the "
-			   "CURRENT CCA master key");
+			   "CURRENT master key");
 
-		rc = select_cca_adapter_by_mkvp(&g.cca, mkvp, NULL,
-						FLAG_SEL_CCA_MATCH_OLD_MKVP,
-						g.verbose);
-		if (rc == -ENOTSUP) {
-			rc = 0;
-			selected = 0;
-		}
+		rc = reencipher_secure_key(&g.lib, secure_key, secure_key_size,
+					   NULL, REENCIPHER_OLD_TO_CURRENT,
+					   &selected, g.verbose);
 		if (rc != 0) {
-			warnx("No APQN found that is suitable for "
-			      "re-enciphering the secure AES volume key");
-			rc = EXIT_FAILURE;
-			goto out;
-		}
-
-		rc = key_token_change(&g.cca, secure_key, secure_key_size,
-				      METHOD_OLD_TO_CURRENT,
-				      g.verbose);
-		if (rc != 0) {
-			warnx("Re-encipher from OLD to CURRENT CCA "
-			      "master key has failed\n");
-			if (!selected)
-				print_msg_for_cca_envvars("secure AES key");
+			if (rc == -ENODEV) {
+				warnx("No APQN found that is suitable for "
+				      "re-enciphering the secure AES volume "
+				      "key");
+			} else {
+				warnx("Re-encipher from OLD to CURRENT "
+				      "master key has failed\n");
+				if (!selected &&
+				    !is_ep11_aes_key(secure_key,
+						     secure_key_size))
+					print_msg_for_cca_envvars(
+							"secure AES key");
+			}
 			rc = EXIT_FAILURE;
 			goto out;
 		}
 	}
 	if (g.tonew) {
 		pr_verbose("Secure key will be re-enciphered from CURRENT "
-			   "to the NEW CCA master key");
+			   "to the NEW master key");
 
-		rc = select_cca_adapter_by_mkvp(&g.cca, mkvp, NULL,
-						FLAG_SEL_CCA_MATCH_CUR_MKVP |
-						FLAG_SEL_CCA_NEW_MUST_BE_SET,
-						g.verbose);
-		if (rc == -ENOTSUP) {
-			rc = 0;
-			selected = 0;
-		}
+		rc = reencipher_secure_key(&g.lib, secure_key, secure_key_size,
+					   NULL, REENCIPHER_CURRENT_TO_NEW,
+					   &selected, g.verbose);
 		if (rc != 0) {
-			util_print_indented("No APQN found that is suitable "
-					    "for re-enciphering this secure "
-					    "AES key and has the NEW master "
-					    "key loaded", 0);
-			rc = EXIT_FAILURE;
-			goto out;
-		}
-
-		rc = key_token_change(&g.cca, secure_key, secure_key_size,
-				      METHOD_CURRENT_TO_NEW, g.verbose);
-		if (rc != 0) {
-			warnx("Re-encipher from CURRENT to NEW CCA "
-			      "master key has failed\n");
-			if (!selected)
-				print_msg_for_cca_envvars("secure AES key");
+			if (rc == -ENODEV) {
+				warnx("No APQN found that is suitable for "
+				      "re-enciphering the secure AES volume "
+				      "key and has the NEW master key loaded");
+			} else {
+				warnx("Re-encipher from CURRENT to NEW "
+				      "master key has failed\n");
+				if (!selected &&
+				    !is_ep11_aes_key(secure_key,
+						     secure_key_size))
+					print_msg_for_cca_envvars(
+							"secure AES key");
+			}
 			rc = EXIT_FAILURE;
 			goto out;
 		}
@@ -1360,7 +1346,7 @@ out:
 /*
  * Command handler for 'reencipher in repository'.
  *
- * Re-encipher the specified secure key with the NEW or CURRENT CCA master key.
+ * Re-encipher the specified secure key with the NEW or CURRENT master key.
  */
 static int command_reencipher_repository(void)
 {
@@ -1395,7 +1381,7 @@ static int command_reencipher_repository(void)
 
 	rc = keystore_reencipher_key(g.keystore, g.name, g.apqns, g.fromold,
 				     g.tonew, g.inplace, g.staged, g.complete,
-				     g.pkey_fd, &g.cca);
+				     g.pkey_fd, &g.lib);
 
 	return rc != 0 ? EXIT_FAILURE : EXIT_SUCCESS;
 }
@@ -1403,7 +1389,7 @@ static int command_reencipher_repository(void)
 /*
  * Command handler for 'reencipher'.
  *
- * Re-encipher the specified secure key with the NEW or CURRENT CCA master key.
+ * Re-encipher the specified secure key with the NEW or CURRENT master key.
  */
 static int command_reencipher(void)
 {
@@ -1425,9 +1411,10 @@ static int command_validate_file(void)
 	char vp[VERIFICATION_PATTERN_LEN];
 	size_t secure_key_size;
 	size_t clear_key_size;
+	const char *key_type;
+	u8 mkvp[MKVP_LENGTH];
 	u8 *secure_key;
 	int is_old_mk;
-	u64 mkvp;
 	int rc;
 
 	if (g.name != NULL) {
@@ -1475,7 +1462,7 @@ static int command_validate_file(void)
 	}
 
 	rc = get_master_key_verification_pattern(secure_key, secure_key_size,
-						 &mkvp, g.verbose);
+						 mkvp, g.verbose);
 	if (rc != 0) {
 		warnx("Failed to get the master key verification pattern: %s",
 		      strerror(-rc));
@@ -1483,24 +1470,27 @@ static int command_validate_file(void)
 		goto out;
 	}
 
+	key_type = get_key_type(secure_key, secure_key_size);
+
 	printf("Validation of secure key in file '%s':\n", g.pos_arg);
 	printf("  Status:                Valid\n");
 	printf("  Secure key size:       %lu bytes\n", secure_key_size);
-	printf("  Key type:              %s\n",
-	       get_key_type(secure_key, secure_key_size));
+	printf("  Key type:              %s\n", key_type);
 	printf("  Clear key size:        %lu bits\n", clear_key_size);
 	printf("  XTS type key:          %s\n",
 	       is_xts_key(secure_key, secure_key_size) ? "Yes" : "No");
-	printf("  Enciphered with:       %s CCA master key (MKVP: %016llx)\n",
-	       is_old_mk ? "OLD" : "CURRENT", mkvp);
+	printf("  Enciphered with:       %s master key (MKVP: %s)\n",
+	       is_old_mk ? "OLD" : "CURRENT",
+	       printable_mkvp(get_card_type_for_keytype(key_type), mkvp));
 	printf("  Verification pattern:  %.*s\n", VERIFICATION_PATTERN_LEN / 2,
 	       vp);
 	printf("                         %.*s\n", VERIFICATION_PATTERN_LEN / 2,
 	       &vp[VERIFICATION_PATTERN_LEN / 2]);
 
 	rc = cross_check_apqns(NULL, mkvp,
-			       get_min_card_level_for_keytype(
-				     get_key_type(secure_key, secure_key_size)),
+			       get_min_card_level_for_keytype(key_type),
+			       get_min_fw_version_for_keytype(key_type),
+			       get_card_type_for_keytype(key_type),
 			       true, g.verbose);
 	if (rc == -EINVAL)
 		return EXIT_FAILURE;
@@ -1571,7 +1561,7 @@ static int command_import(void)
 
 	rc = keystore_import_key(g.keystore, g.name, g.description, g.volumes,
 				 g.apqns, g.noapqncheck, g.sector_size,
-				 g.pos_arg, g.volume_type, &g.cca);
+				 g.pos_arg, g.volume_type, &g.lib);
 
 	return rc != 0 ? EXIT_FAILURE : EXIT_SUCCESS;
 }
@@ -1750,11 +1740,11 @@ static int command_convert_file(void)
 	u8 output_key[2 * MAX_SECURE_KEY_SIZE];
 	unsigned int output_key_size;
 	size_t secure_key_size;
+	u8 mkvp[MKVP_LENGTH];
 	int rc, is_old_mk;
 	int selected = 1;
 	u8 *secure_key;
 	int min_level;
-	u64 mkvp;
 
 	if (g.name != NULL) {
 		warnx("Option '--name|-N' is not valid for "
@@ -1775,7 +1765,10 @@ static int command_convert_file(void)
 		return EXIT_FAILURE;
 	}
 
-	rc = cross_check_apqns(NULL, 0, min_level, true, g.verbose);
+	rc = cross_check_apqns(NULL, NULL, min_level,
+			       get_min_fw_version_for_keytype(g.key_type),
+			       get_card_type_for_keytype(g.key_type),
+			       true, g.verbose);
 	if (rc == -EINVAL)
 		return EXIT_FAILURE;
 	if (rc != 0 && rc != -ENOTSUP) {
@@ -1797,7 +1790,7 @@ static int command_convert_file(void)
 	}
 
 	rc = get_master_key_verification_pattern(secure_key, secure_key_size,
-						 &mkvp, g.verbose);
+						 mkvp, g.verbose);
 	if (rc != 0) {
 		warnx("Failed to get the master key verification pattern: %s",
 		      strerror(-rc));
@@ -1910,7 +1903,7 @@ static int command_convert_repository(void)
 	}
 
 	rc = keystore_convert_key(g.keystore, g.name, g.key_type, g.noapqncheck,
-				  g.force, g.pkey_fd, &g.cca);
+				  g.force, g.pkey_fd, &g.lib);
 
 	return rc != 0 ? EXIT_FAILURE : EXIT_SUCCESS;
 }
@@ -2225,6 +2218,13 @@ int main(int argc, char *argv[])
 			goto out;
 		}
 	}
+	if (command->need_ep11_library) {
+		rc = load_ep11_library(&g.ep11, g.verbose);
+		if (rc != 0) {
+			rc = EXIT_FAILURE;
+			goto out;
+		}
+	}
 	if (command->need_pkey_device) {
 		g.pkey_fd = open_pkey_device(g.verbose);
 		if (g.pkey_fd == -1) {
@@ -2240,6 +2240,8 @@ int main(int argc, char *argv[])
 out:
 	if (g.cca.lib_csulcca)
 		dlclose(g.cca.lib_csulcca);
+	if (g.ep11.lib_ep11)
+		dlclose(g.ep11.lib_ep11);
 	if (g.pkey_fd >= 0)
 		close(g.pkey_fd);
 	if (g.keystore)
